library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgPkg.all;
use work.ffpg_wbgen2_pkg.all;
use work.wishbone_pkg.all;

entity WbSlaveWrapper is
    port (
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        Wb_i: in t_wishbone_slave_in;
        Wb_o: out t_wishbone_slave_out;
        ClkRf_ik: in std_logic;
        Ch1SetMemAddress_ib11: in unsigned(10 downto 0);
        Ch1SetMemData_ob32: out unsigned(31 downto 0);
        Ch1SetMemReadStrobe_i: in std_logic;
        Ch1ResMemAddress_ib11: in unsigned(10 downto 0);
        Ch1ResMemData_ob32: out unsigned(31 downto 0);
        Ch1ResMemReadStrobe_i: in std_logic;
        Ch2SetMemAddress_ib11: in unsigned(10 downto 0);
        Ch2SetMemData_ob32: out unsigned(31 downto 0);
        Ch2SetMemReadStrobe_i: in std_logic;
        Ch2ResMemAddress_ib11: in unsigned(10 downto 0);
        Ch2ResMemData_ob32: out unsigned(31 downto 0);
        Ch2ResMemReadStrobe_i: in std_logic;
        WbRegs_i: in t_ffpg_in_registers;
        WbRegs_o: out t_ffpg_out_registers
    );
end entity;

architecture syn of WbSlaveWrapper is

    signal Reset_nr: std_logic;

    signal WbRegsInput: t_ffpg_in_registers;
    signal WbRegsOutput: t_ffpg_out_registers;

    -- registers for LOAD_EXT fields
    signal control_clock_selection: std_logic_vector(1 downto 0) := (others => '0');
    signal control_ch1_mode: std_logic_vector(1 downto 0) := (others => '0');
    signal control_ch2_mode: std_logic_vector(1 downto 0) := (others => '0');
    signal vcxo_voltage: unsigned(15 downto 0) := (others => '0');
    signal clock_divider_hi: unsigned(3 downto 0) := (others => '0');
    signal ch1_delay_set: unsigned(9 downto 0) := (others => '0');
    signal ch1_delay_reset: unsigned(9 downto 0) := (others => '0');
    signal ch2_delay_set: unsigned(9 downto 0) := (others => '0');
    signal ch2_delay_reset: unsigned(9 downto 0) := (others => '0');
    signal trigger_threshold: unsigned(15 downto 0) := (others => '0');
    signal overflow: unsigned(15 downto 0) := (others => '0');
    signal trigger_latency: unsigned(15 downto 0) := (others => '0');
    -- delayed load signals for LOAD_EXT fields
    signal control_clock_selection_load: std_logic := '0';
    signal control_ch1_mode_load: std_logic := '0';
    signal control_ch2_mode_load: std_logic := '0';
    signal vcxo_voltage_load: std_logic := '0';
    signal clock_divider_hi_load: std_logic := '0';
    signal ch1_delay_set_load: std_logic := '0';
    signal ch1_delay_reset_load: std_logic := '0';
    signal ch2_delay_set_load: std_logic := '0';
    signal ch2_delay_reset_load: std_logic := '0';
    signal trigger_threshold_load: std_logic := '0';
    signal overflow_load: std_logic := '0';
    signal trigger_latency_load: std_logic := '0';
    
begin

    Reset_nr <= not Reset_ir;

    cWbSlave: entity work.ffpg_csr(syn)
        port map (
            rst_n_i => Reset_nr,
            clk_sys_i => Clk_ik,
            wb_adr_i => Wb_i.adr(15 downto 2), -- CSR is 32-bit addressed, WB is byte addressed
            wb_dat_i => Wb_i.dat,
            wb_dat_o => Wb_o.dat,
            wb_cyc_i => Wb_i.cyc,
            wb_sel_i => Wb_i.sel,
            wb_stb_i => Wb_i.stb,
            wb_we_i => Wb_i.we,
            wb_ack_o => Wb_o.ack,
            wb_stall_o => Wb_o.stall,
            clk_rf_ik => ClkRf_ik,
            ffpg_ch1_set_mem_addr_i => std_logic_vector(Ch1SetMemAddress_ib11),
            unsigned(ffpg_ch1_set_mem_data_o) => Ch1SetMemData_ob32,
            ffpg_ch1_set_mem_rd_i => Ch1SetMemReadStrobe_i,
            ffpg_ch1_res_mem_addr_i => std_logic_vector(Ch1ResMemAddress_ib11),
            unsigned(ffpg_ch1_res_mem_data_o) => Ch1ResMemData_ob32,
            ffpg_ch1_res_mem_rd_i => Ch1ResMemReadStrobe_i,
            ffpg_ch2_set_mem_addr_i => std_logic_vector(Ch2SetMemAddress_ib11),
            unsigned(ffpg_ch2_set_mem_data_o) => Ch2SetMemData_ob32,
            ffpg_ch2_set_mem_rd_i => Ch2SetMemReadStrobe_i,
            ffpg_ch2_res_mem_addr_i => std_logic_vector(Ch2ResMemAddress_ib11),
            unsigned(ffpg_ch2_res_mem_data_o) => Ch2ResMemData_ob32,
            ffpg_ch2_res_mem_rd_i => Ch2ResMemReadStrobe_i,
            regs_i => WbRegsInput,
            regs_o => WbRegsOutput
        );
    Wb_o.err <= '0';
    Wb_o.rty <= '0';
    Wb_o.int <= '0';

    -- local registers for LOAD_EXT fields
    pLocalRegs: process (Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if Reset_ir = '1' then
                control_clock_selection_load <= '0';
                control_ch1_mode_load <= '0';
                control_ch2_mode_load <= '0';
                vcxo_voltage_load <= '0';
                clock_divider_hi_load <= '0';
                ch1_delay_set_load <= '0';
                ch1_delay_reset_load <= '0';
                ch2_delay_set_load <= '0';
                ch2_delay_reset_load <= '0';
                trigger_threshold_load <= '0';
                overflow_load <= '0';
                trigger_latency_load <= '0';

                control_clock_selection <= (others => '0');
                control_ch1_mode <= (others => '0');
                control_ch2_mode <= (others => '0');
                vcxo_voltage <= (others => '0');
                clock_divider_hi <= (others => '0');
                ch1_delay_set <= (others => '0');
                ch1_delay_reset <= (others => '0');
                ch2_delay_set <= (others => '0');
                ch2_delay_reset <= (others => '0');
                trigger_threshold <= (others => '0');
                overflow <= (others => '0');
                trigger_latency <= (others => '0');
            else
                control_clock_selection_load <= '0';
                control_ch1_mode_load <= '0';
                control_ch2_mode_load <= '0';
                vcxo_voltage_load <= '0';
                clock_divider_hi_load <= '0';
                ch1_delay_set_load <= '0';
                ch1_delay_reset_load <= '0';
                ch2_delay_set_load <= '0';
                ch2_delay_reset_load <= '0';
                trigger_threshold_load <= '0';
                overflow_load <= '0';
                trigger_latency_load <= '0';

                if WbRegsOutput.control_clock_selection_load_o = '1' then
                    control_clock_selection <= WbRegsOutput.control_clock_selection_o;
                    control_clock_selection_load <= '1';
                end if;
                if WbRegsOutput.control_ch1_mode_load_o = '1' then
                    control_ch1_mode <= WbRegsOutput.control_ch1_mode_o;
                    control_ch1_mode_load <= '1';
                end if;
                if WbRegsOutput.control_ch2_mode_load_o = '1' then
                    control_ch2_mode <= WbRegsOutput.control_ch2_mode_o;
                    control_ch2_mode_load <= '1';
                end if;
                if WbRegsOutput.vcxo_voltage_load_o = '1' then
                    vcxo_voltage <= WbRegsOutput.vcxo_voltage_o;
                    vcxo_voltage_load <= '1';
                end if;
                if WbRegsOutput.clock_divider_hi_load_o = '1' then
                    clock_divider_hi <= WbRegsOutput.clock_divider_hi_o;
                    clock_divider_hi_load <= '1';
                end if;
                if WbRegsOutput.ch1_delay_set_load_o = '1' then
                    ch1_delay_set <= WbRegsOutput.ch1_delay_set_o;
                    ch1_delay_set_load <= '1';
                end if;
                if WbRegsOutput.ch1_delay_reset_load_o = '1' then
                    ch1_delay_reset <= WbRegsOutput.ch1_delay_reset_o;
                    ch1_delay_reset_load <= '1';
                end if;
                if WbRegsOutput.ch2_delay_set_load_o = '1' then
                    ch2_delay_set <= WbRegsOutput.ch2_delay_set_o;
                    ch2_delay_set_load <= '1';
                end if;
                if WbRegsOutput.ch2_delay_reset_load_o = '1' then
                    ch2_delay_reset <= WbRegsOutput.ch2_delay_reset_o;
                    ch2_delay_reset_load <= '1';
                end if;
                if WbRegsOutput.trigger_threshold_load_o = '1' then
                    trigger_threshold <= WbRegsOutput.trigger_threshold_o;
                    trigger_threshold_load <= '1';
                end if;
                if WbRegsOutput.overflow_load_o = '1' then
                    overflow <= WbRegsOutput.overflow_o;
                    overflow_load <= '1';
                end if;
                if WbRegsOutput.trigger_latency_load_o = '1' then
                    trigger_latency <= WbRegsOutput.trigger_latency_o;
                    trigger_latency_load <= '1';
                end if;
            end if;
        end if;
    end process;
    
    pInputRegisters: process (WbRegs_i, control_clock_selection, control_ch1_mode, control_ch2_mode, vcxo_voltage, clock_divider_hi, ch1_delay_set, ch1_delay_reset, ch2_delay_set, ch2_delay_reset, trigger_threshold, overflow, trigger_latency) is begin
        -- #!@& ISE doesn't know VHDL 2008 (... process (all) ...)
        -- by default, all values are passed
        WbRegsInput <= WbRegs_i;
        -- LOAD_EXT inputs are overwritten by local registers
        WbRegsInput.control_clock_selection_i <= control_clock_selection;
        WbRegsInput.control_ch1_mode_i <= control_ch1_mode;
        WbRegsInput.control_ch2_mode_i <= control_ch2_mode;
        WbRegsInput.vcxo_voltage_i <= vcxo_voltage;
        WbRegsInput.clock_divider_hi_i <= clock_divider_hi;
        WbRegsInput.ch1_delay_set_i <= ch1_delay_set;
        WbRegsInput.ch1_delay_reset_i <= ch1_delay_reset;
        WbRegsInput.ch2_delay_set_i <= ch2_delay_set;
        WbRegsInput.ch2_delay_reset_i <= ch2_delay_reset;
        WbRegsInput.trigger_threshold_i <= trigger_threshold;
        WbRegsInput.overflow_i <= overflow;
        WbRegsInput.trigger_latency_i <= trigger_latency;
    end process;
    
    pOutputRegisters: process (WbRegsOutput, control_clock_selection, control_ch1_mode, control_ch2_mode, vcxo_voltage, clock_divider_hi, ch1_delay_set, ch1_delay_reset, ch2_delay_set, ch2_delay_reset, trigger_threshold, overflow, trigger_latency, control_clock_selection_load, control_ch1_mode_load, control_ch2_mode_load, vcxo_voltage_load, clock_divider_hi_load, ch1_delay_set_load, ch1_delay_reset_load, ch2_delay_set_load, ch2_delay_reset_load, trigger_threshold_load, overflow_load, trigger_latency_load) is begin
        -- #!@& ISE doesn't know VHDL 2008 (... process(all) ...)
        -- by default, all values are passed
        WbRegs_o <= WbRegsOutput;
        -- LOAD_EXT inputs are overwritten by local registers
        WbRegs_o.control_clock_selection_o <= control_clock_selection;
        WbRegs_o.control_ch1_mode_o <= control_ch1_mode;
        WbRegs_o.control_ch2_mode_o <= control_ch2_mode;
        WbRegs_o.vcxo_voltage_o <= vcxo_voltage;
        WbRegs_o.clock_divider_hi_o <= clock_divider_hi;
        WbRegs_o.ch1_delay_set_o <= ch1_delay_set;
        WbRegs_o.ch1_delay_reset_o <= ch1_delay_reset;
        WbRegs_o.ch2_delay_set_o <= ch2_delay_set;
        WbRegs_o.ch2_delay_reset_o <= ch2_delay_reset;
        WbRegs_o.trigger_threshold_o <= trigger_threshold;
        WbRegs_o.overflow_o <= overflow;
        WbRegs_o.trigger_latency_o <= trigger_latency;
        WbRegs_o.control_clock_selection_load_o <= control_clock_selection_load;
        WbRegs_o.control_ch1_mode_load_o <= control_ch1_mode_load;
        WbRegs_o.control_ch2_mode_load_o <= control_ch2_mode_load;
        WbRegs_o.vcxo_voltage_load_o <= vcxo_voltage_load;
        WbRegs_o.clock_divider_hi_load_o <= clock_divider_hi_load;
        WbRegs_o.ch1_delay_set_load_o <= ch1_delay_set_load;
        WbRegs_o.ch1_delay_reset_load_o <= ch1_delay_reset_load;
        WbRegs_o.ch2_delay_set_load_o <= ch2_delay_set_load;
        WbRegs_o.ch2_delay_reset_load_o <= ch2_delay_reset_load;
        WbRegs_o.trigger_threshold_load_o <= trigger_threshold_load;
        WbRegs_o.overflow_load_o <= overflow_load;
        WbRegs_o.trigger_latency_load_o <= trigger_latency_load;
    end process;

end architecture;